#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    check_optimal_termination,
    units as pyunits,
)

from pyomo.util.check_units import assert_units_consistent
from idaes.core import (
    FlowsheetBlock,
    UnitModelCostingBlock,
)
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
)
from idaes.core.util.scaling import (
    calculate_scaling_factors,
)
from idaes.core.util.testing import initialization_tester

from watertap.property_models.multicomp_aq_sol_prop_pack import (
    MCASParameterBlock,
    ActivityCoefficientModel,
    DensityCalculation,
    MaterialFlowBasis,
)

from watertap.unit_models.stoichiometric_reactor import StoichiometricReactor
from watertap.costing import WaterTAPCosting

__author__ = "Timothy Bartholomew, Alexander Dudchenko"

solver = get_solver()


class TestStoichiometricReactor:
    @pytest.fixture(scope="class")
    def basic_unit_mass(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        default_dict = {
            "solute_list": [
                "Ca_2+",
                "HCO3_-",
                "Na_+",
                "Cl_-",
                "Mg_2+",
            ],
            "diffusivity_data": {
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "HCO3_-"): 1.19e-9,
                ("Liq", "Na_+"): 1.33e-9,
                ("Liq", "Cl_-"): 2.03e-9,
                ("Liq", "Mg_2+"): 0.706e-9,
            },
            "mw_data": {
                "H2O": 18e-3,
                "Ca_2+": 40e-3,
                "HCO3_-": 61.0168e-3,
                "Na_+": 23e-3,
                "Cl_-": 35e-3,
                "Mg_2+": 24.305e-3,
            },
            "stokes_radius_data": {
                "Ca_2+": 0.309e-9,
                "HCO3_-": 2.06e-10,
                "Cl_-": 0.121e-9,
                "Na_+": 0.184e-9,
                "Mg_2+": 0.347e-9,
            },
            "charge": {
                "Ca_2+": 2,
                "HCO3_-": -1,
                "Na_+": 1,
                "Cl_-": -1,
                "Mg_2+": 2,
            },
            "activity_coefficient_model": ActivityCoefficientModel.ideal,
            "density_calculation": DensityCalculation.constant,
            "material_flow_basis": MaterialFlowBasis.mass,
        }
        m.fs.properties = MCASParameterBlock(**default_dict)
        # Define precipitatnts
        precipitants = {
            "Calcite": {
                "mw": 100.09 * pyunits.g / pyunits.mol,
                "precipitation_stoichiometric": {"Ca_2+": 1, "HCO3_-": 1},
            },
            "Brucite": {
                "mw": 58.3197 * pyunits.g / pyunits.mol,
                "precipitation_stoichiometric": {"Mg_2+": 1, "H2O": 1},
            },
        }
        # Define reagents
        reagents = {
            "Na2CO3": {
                "mw": 105.99 * pyunits.g / pyunits.mol,
                "dissolution_stoichiometric": {"Na_+": 2, "HCO3_-": 1},
            },
            "CaO": {
                "mw": 56.0774 * pyunits.g / pyunits.mol,
                "dissolution_stoichiometric": {"Ca_2+": 1, "H2O": 1},
            },
        }
        m.fs.unit = StoichiometricReactor(
            property_package=m.fs.properties,
            reagent=reagents,
            precipitate=precipitants,
        )

        m.fs.unit.waste_mass_frac_precipitate.fix(0.2)
        # the reactor us assumed performance model so we are simply testing that
        # mass balances are correct
        m.fs.unit.reagent_dose["Na2CO3"].fix(1e-3)
        m.fs.unit.reagent_dose["CaO"].fix(1e-3)

        m.fs.unit.flow_mass_precipitate["Calcite"].fix(1e-3)
        m.fs.unit.flow_mass_precipitate["Brucite"].fix(1e-4)

        # fix_fix_comp
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(1)
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Ca_2+"].fix(0.001)
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "HCO3_-"].fix(0.001)
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Na_+"].fix(0.001)
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Cl_-"].fix(0.001)
        m.fs.unit.inlet.flow_mass_phase_comp[0, "Liq", "Mg_2+"].fix(0.001)

        for key, obj in m.fs.unit.inlet.flow_mass_phase_comp.items():
            val = value(obj)
            m.fs.properties.set_default_scaling(
                "flow_mass_phase_comp", 1 / val, index=("Liq", key[-1])
            )

        m.fs.unit.inlet.pressure[0].fix(101325)
        m.fs.unit.inlet.temperature[0].fix(273.15 + 20)

        return m

    @pytest.fixture(scope="class")
    def basic_unit_molar(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        default_dict = {
            "solute_list": [
                "Ca_2+",
                "HCO3_-",
                "Na_+",
                "Cl_-",
                "Mg_2+",
            ],
            "diffusivity_data": {
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "HCO3_-"): 1.19e-9,
                ("Liq", "Na_+"): 1.33e-9,
                ("Liq", "Cl_-"): 2.03e-9,
                ("Liq", "Mg_2+"): 0.706e-9,
            },
            "mw_data": {
                "H2O": 18e-3,
                "Ca_2+": 40e-3,
                "HCO3_-": 61.0168e-3,
                "Na_+": 23e-3,
                "Cl_-": 35e-3,
                "Mg_2+": 24.305e-3,
            },
            "stokes_radius_data": {
                "Ca_2+": 0.309e-9,
                "HCO3_-": 2.06e-10,
                "Cl_-": 0.121e-9,
                "Na_+": 0.184e-9,
                "Mg_2+": 0.347e-9,
            },
            "charge": {
                "Ca_2+": 2,
                "HCO3_-": -1,
                "Na_+": 1,
                "Cl_-": -1,
                "Mg_2+": 2,
            },
            "activity_coefficient_model": ActivityCoefficientModel.ideal,
            "density_calculation": DensityCalculation.constant,
            "material_flow_basis": MaterialFlowBasis.molar,
        }
        m.fs.properties = MCASParameterBlock(**default_dict)
        # Define precipitatnts
        precipitants = {
            "Calcite": {
                "mw": 100.09 * pyunits.g / pyunits.mol,
                "precipitation_stoichiometric": {"Ca_2+": 1, "HCO3_-": 1},
            },
            "Brucite": {
                "mw": 58.3197 * pyunits.g / pyunits.mol,
                "precipitation_stoichiometric": {"Mg_2+": 1, "H2O": 1},
            },
        }
        # Define reagents
        reagents = {
            "Na2CO3": {
                "mw": 105.99 * pyunits.g / pyunits.mol,
                "dissolution_stoichiometric": {"Na_+": 2, "HCO3_-": 1},
            },
            "CaO": {
                "mw": 56.0774 * pyunits.g / pyunits.mol,
                "dissolution_stoichiometric": {"Ca_2+": 1, "H2O": 1},
            },
        }
        m.fs.unit = StoichiometricReactor(
            property_package=m.fs.properties,
            reagent=reagents,
            precipitate=precipitants,
        )

        m.fs.unit.waste_mass_frac_precipitate.fix(0.2)
        # the reactor us assumed performance model so we are simply testing that
        # mass balances are correct
        m.fs.unit.reagent_dose["Na2CO3"].fix(1e-3)
        m.fs.unit.reagent_dose["CaO"].fix(1e-3)

        m.fs.unit.flow_mass_precipitate["Calcite"].fix(1e-3)
        m.fs.unit.flow_mass_precipitate["Brucite"].fix(1e-4)

        # fix_fix_comp
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(55)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "HCO3_-"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Mg_2+"].fix(0.1)

        for key, obj in m.fs.unit.inlet.flow_mol_phase_comp.items():
            val = value(obj)
            m.fs.properties.set_default_scaling(
                "flow_mol_phase_comp", 1 / val, index=("Liq", key[-1])
            )

        m.fs.unit.inlet.pressure[0].fix(101325)
        m.fs.unit.inlet.temperature[0].fix(273.15 + 20)

        return m

    @pytest.fixture(scope="class")
    def dissolution_reactor(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        default_dict = {
            "solute_list": [
                "Ca_2+",
                "HCO3_-",
                "Na_+",
                "Cl_-",
                "Mg_2+",
            ],
            "diffusivity_data": {
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "HCO3_-"): 1.19e-9,
                ("Liq", "Na_+"): 1.33e-9,
                ("Liq", "Cl_-"): 2.03e-9,
                ("Liq", "Mg_2+"): 0.706e-9,
            },
            "mw_data": {
                "H2O": 18e-3,
                "Ca_2+": 40e-3,
                "HCO3_-": 61.0168e-3,
                "Na_+": 23e-3,
                "Cl_-": 35e-3,
                "Mg_2+": 24.305e-3,
            },
            "stokes_radius_data": {
                "Ca_2+": 0.309e-9,
                "HCO3_-": 2.06e-10,
                "Cl_-": 0.121e-9,
                "Na_+": 0.184e-9,
                "Mg_2+": 0.347e-9,
            },
            "charge": {
                "Ca_2+": 2,
                "HCO3_-": -1,
                "Na_+": 1,
                "Cl_-": -1,
                "Mg_2+": 2,
            },
            "activity_coefficient_model": ActivityCoefficientModel.ideal,
            "density_calculation": DensityCalculation.constant,
            "material_flow_basis": MaterialFlowBasis.molar,
        }
        m.fs.properties = MCASParameterBlock(**default_dict)
        # Define reagents
        reagents = {
            "Na2CO3": {
                "mw": 105.99 * pyunits.g / pyunits.mol,
                "dissolution_stoichiometric": {"Na_+": 2, "HCO3_-": 1},
                "reagent_density": 1.2 * pyunits.kg / pyunits.L,
            },
            "CaO": {
                "mw": 56.0774 * pyunits.g / pyunits.mol,
                "dissolution_stoichiometric": {"Ca_2+": 1, "H2O": 1},
                "reagent_density": 1.2 * pyunits.kg / pyunits.L,
            },
        }
        m.fs.unit = StoichiometricReactor(
            property_package=m.fs.properties,
            reagent=reagents,
        )

        # the reactor us assumed performance model so we are simply testing that
        # mass balances are correct
        m.fs.unit.reagent_dose["Na2CO3"].fix(1e-3)
        m.fs.unit.reagent_dose["CaO"].fix(1e-3)
        # fix_fix_comp
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(55)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "HCO3_-"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Mg_2+"].fix(0.1)

        for key, obj in m.fs.unit.inlet.flow_mol_phase_comp.items():
            val = value(obj)
            m.fs.properties.set_default_scaling(
                "flow_mol_phase_comp", 1 / val, index=("Liq", key[-1])
            )

        m.fs.unit.inlet.pressure[0].fix(101325)
        m.fs.unit.inlet.temperature[0].fix(273.15 + 20)

        return m

    @pytest.fixture(scope="class")
    def precipitation_reactor(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        default_dict = {
            "solute_list": [
                "Ca_2+",
                "HCO3_-",
                "Na_+",
                "Cl_-",
                "Mg_2+",
            ],
            "diffusivity_data": {
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "HCO3_-"): 1.19e-9,
                ("Liq", "Na_+"): 1.33e-9,
                ("Liq", "Cl_-"): 2.03e-9,
                ("Liq", "Mg_2+"): 0.706e-9,
            },
            "mw_data": {
                "H2O": 18e-3,
                "Ca_2+": 40e-3,
                "HCO3_-": 61.0168e-3,
                "Na_+": 23e-3,
                "Cl_-": 35e-3,
                "Mg_2+": 24.305e-3,
            },
            "stokes_radius_data": {
                "Ca_2+": 0.309e-9,
                "HCO3_-": 2.06e-10,
                "Cl_-": 0.121e-9,
                "Na_+": 0.184e-9,
                "Mg_2+": 0.347e-9,
            },
            "charge": {
                "Ca_2+": 2,
                "HCO3_-": -1,
                "Na_+": 1,
                "Cl_-": -1,
                "Mg_2+": 2,
            },
            "activity_coefficient_model": ActivityCoefficientModel.ideal,
            "density_calculation": DensityCalculation.constant,
            "material_flow_basis": MaterialFlowBasis.molar,
        }
        m.fs.properties = MCASParameterBlock(**default_dict)
        # Define reagents
        # Define reagents
        precipitants = {
            "Calcite": {
                "mw": 100.09 * pyunits.g / pyunits.mol,
                "precipitation_stoichiometric": {"Ca_2+": 1, "HCO3_-": 1},
            },
            "Brucite": {
                "mw": 58.3197 * pyunits.g / pyunits.mol,
                "precipitation_stoichiometric": {"Mg_2+": 1, "H2O": 1},
            },
        }
        m.fs.unit = StoichiometricReactor(
            property_package=m.fs.properties,
            precipitate=precipitants,
        )

        m.fs.unit.waste_mass_frac_precipitate.fix(0.2)
        # the reactor us assumed performance model so we are simply testing that
        # mass balances are correct

        m.fs.unit.flow_mass_precipitate["Calcite"].fix(1e-3)
        m.fs.unit.flow_mass_precipitate["Brucite"].fix(1e-4)
        m.fs.unit.waste_mass_frac_precipitate.fix(0.2)

        # fix_fix_comp
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "H2O"].fix(55)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Ca_2+"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "HCO3_-"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Na_+"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Cl_-"].fix(0.1)
        m.fs.unit.inlet.flow_mol_phase_comp[0, "Liq", "Mg_2+"].fix(0.1)

        for key, obj in m.fs.unit.inlet.flow_mol_phase_comp.items():
            val = value(obj)
            m.fs.properties.set_default_scaling(
                "flow_mol_phase_comp", 1 / val, index=("Liq", key[-1])
            )

        m.fs.unit.inlet.pressure[0].fix(101325)
        m.fs.unit.inlet.temperature[0].fix(273.15 + 20)

        return m

    @pytest.mark.unit
    def test_dof_units(self, basic_unit_molar, basic_unit_mass):
        m = basic_unit_molar
        assert degrees_of_freedom(m) == 0
        assert assert_units_consistent(m) is None
        m = basic_unit_mass
        assert degrees_of_freedom(m) == 0
        assert assert_units_consistent(m) is None

    @pytest.mark.unit
    def test_molar(self, basic_unit_molar):
        m = basic_unit_molar
        calculate_scaling_factors(m)

        initialization_tester(m)

        flow_ca_in_reagent = (
            1e-3
            / (56.0774 * 1e-3)
            * value(
                m.fs.unit.dissolution_reactor.properties_in[0].flow_vol_phase["Liq"]
            )
        )
        expected_ca_mol_flow = 0.1 + flow_ca_in_reagent

        assert pytest.approx(flow_ca_in_reagent, rel=1e-5) == value(
            m.fs.unit.flow_mass_reagent["CaO"] / (56.0774 * 1e-3)
        )

        assert pytest.approx(flow_ca_in_reagent, rel=1e-5) == value(
            m.fs.unit.flow_mol_reagent["CaO"]
        )

        assert pytest.approx(expected_ca_mol_flow, rel=1e-5) == value(
            m.fs.unit.dissolution_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Ca_2+"
            ]
        )

        expected_ca_mol_flow = 0.1 + flow_ca_in_reagent - 1e-3 / (100.09 * 1e-3)
        assert pytest.approx(expected_ca_mol_flow, rel=1e-5) == value(
            m.fs.unit.precipitation_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Ca_2+"
            ]
        )

        flow_na_in_reagent = (
            1e-3
            / (105.99 * 1e-3)
            * value(
                m.fs.unit.dissolution_reactor.properties_in[0].flow_vol_phase["Liq"]
            )
        )

        expected_na_mol_flow = 0.1 + flow_na_in_reagent * 2

        assert pytest.approx(flow_na_in_reagent, rel=1e-5) == value(
            m.fs.unit.flow_mass_reagent["Na2CO3"] / (105.99 * 1e-3)
        )

        assert pytest.approx(expected_na_mol_flow, rel=1e-5) == value(
            m.fs.unit.dissolution_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Na_+"
            ]
        )

        expected_mg_mol_flow = 0.1 - 1e-4 / (58.3197 * 1e-3)

        assert pytest.approx(expected_mg_mol_flow, rel=1e-5) == value(
            m.fs.unit.precipitation_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Mg_2+"
            ]
        )

    @pytest.mark.unit
    def test_mass(self, basic_unit_mass):
        m = basic_unit_mass
        calculate_scaling_factors(m)

        initialization_tester(m)

        flow_ca_in_reagent = 1e-3 * value(
            m.fs.unit.dissolution_reactor.properties_in[0].flow_vol_phase["Liq"]
        )
        expected_ca_flow = (
            0.001 + flow_ca_in_reagent / 56.0774 * 40
        )  # - 1e-3 / 100.9 * 40

        assert pytest.approx(flow_ca_in_reagent, rel=1e-2) == value(
            m.fs.unit.flow_mass_reagent["CaO"]
        )

        assert pytest.approx(expected_ca_flow, rel=1e-2) == value(
            m.fs.unit.dissolution_reactor.properties_out[0].flow_mass_phase_comp[
                "Liq", "Ca_2+"
            ]
        )
        expected_ca_flow = (
            0.001 + flow_ca_in_reagent / 56.0774 * 40
        ) - 1e-3 / 100.9 * 40
        assert pytest.approx(expected_ca_flow, rel=1e-2) == value(
            m.fs.unit.precipitation_reactor.properties_out[0].flow_mass_phase_comp[
                "Liq", "Ca_2+"
            ]
        )
        flow_na_in_reagent = 1e-3 * value(
            m.fs.unit.dissolution_reactor.properties_in[0].flow_vol_phase["Liq"]
        )

        expected_na_flow = 0.001 + flow_na_in_reagent / 105.99 * 2 * 23

        assert pytest.approx(flow_na_in_reagent, rel=1e-5) == value(
            m.fs.unit.flow_mass_reagent["Na2CO3"]
        )

        assert pytest.approx(expected_na_flow, rel=1e-5) == value(
            m.fs.unit.dissolution_reactor.properties_out[0].flow_mass_phase_comp[
                "Liq", "Na_+"
            ]
        )

        expected_mg_flow = 0.001 - 1e-4 / 58.3197 * 24.305

        assert pytest.approx(expected_mg_flow, rel=1e-5) == value(
            m.fs.unit.precipitation_reactor.properties_out[0].flow_mass_phase_comp[
                "Liq", "Mg_2+"
            ]
        )

    @pytest.mark.unit
    def test_dissolution(self, dissolution_reactor):
        m = dissolution_reactor
        calculate_scaling_factors(m)

        initialization_tester(m)

        flow_ca_in_reagent = (
            1e-3
            / (56.0774 * 1e-3)
            * value(
                m.fs.unit.dissolution_reactor.properties_in[0].flow_vol_phase["Liq"]
            )
        )
        expected_ca_mol_flow = 0.1 + flow_ca_in_reagent

        assert pytest.approx(flow_ca_in_reagent, rel=1e-5) == value(
            m.fs.unit.flow_mass_reagent["CaO"] / (56.0774 * 1e-3)
        )

        assert pytest.approx(expected_ca_mol_flow, rel=1e-5) == value(
            m.fs.unit.dissolution_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Ca_2+"
            ]
        )

        flow_na_in_reagent = (
            1e-3
            / (105.99 * 1e-3)
            * value(
                m.fs.unit.dissolution_reactor.properties_in[0].flow_vol_phase["Liq"]
            )
        )

        expected_na_mol_flow = 0.1 + flow_na_in_reagent * 2

        assert pytest.approx(flow_na_in_reagent, rel=1e-5) == value(
            m.fs.unit.flow_mass_reagent["Na2CO3"] / (105.99 * 1e-3)
        )

        assert pytest.approx(expected_na_mol_flow, rel=1e-5) == value(
            m.fs.unit.dissolution_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Na_+"
            ]
        )

    @pytest.mark.unit
    def test_precipitation(self, precipitation_reactor):
        m = precipitation_reactor
        calculate_scaling_factors(m)

        initialization_tester(m)

        expected_ca_mol_flow = 0.1 - 1e-3 / (100.09 * 1e-3)
        assert pytest.approx(expected_ca_mol_flow, rel=1e-5) == value(
            m.fs.unit.precipitation_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Ca_2+"
            ]
        )
        assert pytest.approx(1e-3 / (100.09 * 1e-3), rel=1e-5) == value(
            m.fs.unit.flow_mol_precipitate["Calcite"]
        )
        expected_mg_mol_flow = 0.1 - 1e-4 / (58.3197 * 1e-3)

        assert pytest.approx(expected_mg_mol_flow, rel=1e-5) == value(
            m.fs.unit.precipitation_reactor.properties_out[0].flow_mol_phase_comp[
                "Liq", "Mg_2+"
            ]
        )

    @pytest.mark.unit
    def test_failed_build(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        default_dict = {
            "solute_list": [
                "Ca_2+",
                "HCO3_-",
                "Na_+",
                "Cl_-",
                "Mg_2+",
            ],
            "diffusivity_data": {
                ("Liq", "Ca_2+"): 9.2e-10,
                ("Liq", "HCO3_-"): 1.19e-9,
                ("Liq", "Na_+"): 1.33e-9,
                ("Liq", "Cl_-"): 2.03e-9,
                ("Liq", "Mg_2+"): 0.706e-9,
            },
            "mw_data": {
                "H2O": 18e-3,
                "Ca_2+": 40e-3,
                "HCO3_-": 61.0168e-3,
                "Na_+": 23e-3,
                "Cl_-": 35e-3,
                "Mg_2+": 24.305e-3,
            },
            "stokes_radius_data": {
                "Ca_2+": 0.309e-9,
                "HCO3_-": 2.06e-10,
                "Cl_-": 0.121e-9,
                "Na_+": 0.184e-9,
                "Mg_2+": 0.347e-9,
            },
            "charge": {
                "Ca_2+": 2,
                "HCO3_-": -1,
                "Na_+": 1,
                "Cl_-": -1,
                "Mg_2+": 2,
            },
            "activity_coefficient_model": ActivityCoefficientModel.ideal,
            "density_calculation": DensityCalculation.constant,
            "material_flow_basis": MaterialFlowBasis.molar,
        }
        m.fs.properties = MCASParameterBlock(**default_dict)
        with pytest.raises(TypeError):
            m.fs.unit = StoichiometricReactor(
                property_package=m.fs.properties,
            )

    @pytest.mark.unit
    def test_costing_softening(self, basic_unit_molar):
        m = basic_unit_molar
        m.fs.costing = WaterTAPCosting()
        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.initialize()
        calculate_scaling_factors(m)
        result = solver.solve(m)
        assert check_optimal_termination(result)
        assert assert_units_consistent(m) is None
        assert degrees_of_freedom(m) == 0

        assert (
            value(m.fs.costing.stoichiometric_reactor.capital_cost_softening) == 374.9
        )
        assert pytest.approx(value(m.fs.costing.total_capital_cost), rel=1e-3) == 245.34
        assert pytest.approx(value(m.fs.costing.total_operating_cost), rel=1e-3) == 7.36

    @pytest.mark.unit
    def test_costing_acid_addition(self, dissolution_reactor):
        # NOTE: testing costing for a dissultion reactor only - we are useing
        # soda ash adn lime addition, but costing is for HCl addition
        # just a test
        m = dissolution_reactor
        m.fs.costing = WaterTAPCosting()
        m.fs.unit.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)
        m.fs.costing.cost_process()
        m.fs.costing.initialize()
        calculate_scaling_factors(m)
        result = solver.solve(m)
        assert check_optimal_termination(result)
        assert assert_units_consistent(m) is None
        assert degrees_of_freedom(m) == 0

        assert (
            value(m.fs.costing.stoichiometric_reactor.capital_cost_acid_addition)
            == 127.8
        )
        assert pytest.approx(value(m.fs.costing.total_capital_cost), rel=1e-3) == 10.02
        assert (
            pytest.approx(value(m.fs.costing.total_operating_cost), rel=1e-3) == 0.30065
        )



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    Var,
)
from idaes.core import FlowsheetBlock
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from pyomo.util.check_units import assert_units_consistent
from idaes.models.unit_models import Heater, Separator, Mixer, Product, Feed

from idaes.models.unit_models.heat_exchanger import (
    HeatExchanger,
)
from watertap.unit_models.pressure_changer import Pump
from watertap.unit_models.MD.membrane_distillation_0D import MembraneDistillation0D
import watertap.property_models.seawater_prop_pack as props_sw
import watertap.property_models.water_prop_pack as props_w
from watertap.flowsheets.MD.MD_single_stage_continuous_recirculation import (
    main,
    build,
    set_operating_conditions,
    optimize_set_up,
    initialize_system,
    solve,
)

solver = get_solver()


# -----------------------------------------------------------------------------
class TestMDContinuousRecirculation:
    @pytest.fixture(scope="class")
    def system_frame(self):
        m = build()

        return m

    @pytest.mark.unit
    def test_build(self, system_frame):
        m = system_frame
        # model set up
        assert isinstance(m, ConcreteModel)
        assert isinstance(m.fs, FlowsheetBlock)
        assert isinstance(m.fs.properties_hot_ch, props_sw.SeawaterParameterBlock)
        assert isinstance(m.fs.properties_cold_ch, props_sw.SeawaterParameterBlock)
        assert isinstance(m.fs.properties_vapor, props_w.WaterParameterBlock)

        # unit models
        assert isinstance(m.fs.feed, Feed)
        assert isinstance(m.fs.pump_feed, Pump)
        assert isinstance(m.fs.pump_permeate, Pump)
        assert isinstance(m.fs.pump_brine, Pump)
        assert isinstance(m.fs.mixer, Mixer)
        assert isinstance(m.fs.heater, Heater)
        assert isinstance(m.fs.chiller, Heater)
        assert isinstance(m.fs.hx, HeatExchanger)
        assert isinstance(m.fs.MD, MembraneDistillation0D)
        assert isinstance(m.fs.separator_concentrate, Separator)
        assert isinstance(m.fs.separator_permeate, Separator)
        assert isinstance(m.fs.permeate, Product)

        # additional variables
        assert isinstance(m.fs.overall_recovery, Var)
        assert isinstance(m.fs.recycle_ratio, Var)

        # arcs
        arc_dict = {
            # brine (MD hot side) loop
            m.fs.s01: (m.fs.feed.outlet, m.fs.pump_feed.inlet),
            m.fs.s02: (m.fs.pump_feed.outlet, m.fs.mixer.feed),
            m.fs.s03: (m.fs.mixer.outlet, m.fs.hx.cold_inlet),
            m.fs.s04: (m.fs.hx.cold_outlet, m.fs.pump_brine.inlet),
            m.fs.s05: (m.fs.pump_brine.outlet, m.fs.heater.inlet),
            m.fs.s06: (m.fs.heater.outlet, m.fs.MD.hot_ch_inlet),
            m.fs.s07: (m.fs.MD.hot_ch_outlet, m.fs.separator_concentrate.inlet),
            m.fs.s08: (m.fs.separator_concentrate.reject, m.fs.reject.inlet),
            m.fs.s09: (m.fs.separator_concentrate.recycle, m.fs.mixer.recycle),
            # pemeate (MD cold side) loop
            m.fs.s10: (m.fs.chiller.outlet, m.fs.MD.cold_ch_inlet),
            m.fs.s11: (m.fs.MD.cold_ch_outlet, m.fs.hx.hot_inlet),
            m.fs.s12: (m.fs.hx.hot_outlet, m.fs.separator_permeate.inlet),
            m.fs.s13: (m.fs.separator_permeate.permeate, m.fs.permeate.inlet),
            m.fs.s14: (
                m.fs.separator_permeate.cold_loop_stream,
                m.fs.pump_permeate.inlet,
            ),
            m.fs.s15: (m.fs.pump_permeate.outlet, m.fs.chiller.inlet),
        }
        for arc, port_tpl in arc_dict.items():
            assert arc.source is port_tpl[0]
            assert arc.destination is port_tpl[1]

    @pytest.mark.component
    def test_units(self, system_frame):
        assert_units_consistent(system_frame)

    @pytest.mark.component
    def test_set_operating_conditions(self, system_frame):
        m = system_frame
        set_operating_conditions(m)
        assert degrees_of_freedom(m) == 0

    @pytest.mark.component
    @pytest.mark.requires_idaes_solver
    def test_initialize_system(self, system_frame):
        m = system_frame
        initialize_system(m, solver=solver)
        assert degrees_of_freedom(m) == 0

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_solution(self, system_frame):
        m = system_frame
        optimize_set_up(m)
        solve(m, solver=solver)
        assert pytest.approx(0, rel=1e-5) == value(
            m.fs.permeate.flow_mass_phase_comp[0, "Liq", "TDS"]
        )
        assert pytest.approx(0.5, rel=1e-5) == value(m.fs.overall_recovery)
        assert pytest.approx(6.169, rel=1e-3) == value(m.fs.recycle_ratio[0])
        assert pytest.approx(15.44, rel=1e-3) == value(m.fs.costing.LCOW)
        assert pytest.approx(182.9, rel=1e-3) == value(
            m.fs.costing.specific_energy_consumption
        )

    @pytest.mark.requires_idaes_solver
    @pytest.mark.component
    def test_main(self):
        main()



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from pyomo.environ import (
    NegativeReals,
    Set,
    Var,
)
from idaes.core import (
    declare_process_block_class,
    FlowDirection,
)
from idaes.core.util import scaling as iscale
from idaes.core.util.exceptions import ConfigurationError
from idaes.core.util.misc import add_object_reference
from idaes.core.base.control_volume1d import ControlVolume1DBlockData
import idaes.logger as idaeslog
from .MD_channel_base import (
    MDChannelMixin,
    PressureChangeType,
)

__author__ = "Elmira Shamlou"


@declare_process_block_class("MDChannel1DBlock")
class MDChannel1DBlockData(MDChannelMixin, ControlVolume1DBlockData):
    def _skip_element(self, x):
        if self.config.transformation_scheme != "FORWARD":
            return x == self.length_domain.first()
        else:
            return x == self.length_domain.last()

    def apply_transformation(self, *args, **kwargs):
        super().apply_transformation(*args, **kwargs)
        self.difference_elements = Set(
            ordered=True,
            initialize=(x for x in self.length_domain if not self._skip_element(x)),
        )
        self._set_nfe()

    def add_geometry(
        self, length_var, width_var, flow_direction=FlowDirection.forward, **kwargs
    ):
        """
        Method to create spatial domain and volume Var in ControlVolume.

        Args:
            length_var - An external variable to use for the length of
                         the channel. If a variable is provided, a
                         reference will be made to this in place of the length
                         Var.

            width_var - An external variable to use for the width of
                         the channel. If a variable is provided, a
                         reference will be made to this in place of the length
                         Var.
            flow_direction - argument indicating direction of material flow
                            relative to length domain. Valid values:
                                - FlowDirection.forward (default), flow goes
                                  from 0 to 1.
                                - FlowDirection.backward, flow goes from 1 to 0
            length_domain - (optional) a ContinuousSet to use as the length
                            domain for the ControlVolume. If not provided, a
                            new ContinuousSet will be created (default=None).
                            ContinuousSet should be normalized to run between
                            0 and 1.
            length_domain_set - (optional) list of point to use to initialize
                            a new ContinuousSet if length_domain is not
                            provided (default = [0.0, 1.0]).

        Returns:
            None
        """
        super().add_geometry(
            length_var=length_var, flow_direction=flow_direction, **kwargs
        )
        add_object_reference(self, "width", width_var)

    def _add_pressure_change(self, pressure_change_type=PressureChangeType.calculated):
        add_object_reference(self, "dP_dx", self.deltaP)

        units_meta = self.config.property_package.get_metadata().get_derived_units
        self.deltaP_channel = Var(
            self.flowsheet().config.time,
            initialize=-1e5,
            bounds=(-1e6, 0),
            domain=NegativeReals,
            units=units_meta("pressure"),
            doc="total prossure drop across the channel",
        )

    def _add_deltaP(self, pressure_change_type=PressureChangeType.calculated):

        if pressure_change_type == PressureChangeType.fixed_per_stage:

            @self.Constraint(
                self.flowsheet().config.time,
                self.length_domain,
                doc="pressure change due to friction",
            )
            def eq_pressure_change(b, t, x):
                return b.deltaP_channel[t] == b.dP_dx[t, x] * b.length

        else:

            @self.Constraint(
                self.flowsheet().config.time, doc="Total Pressure drop across channel"
            )
            def eq_pressure_change(b, t):
                return b.deltaP_channel[t] == sum(
                    b.dP_dx[t, x] * b.length / b.nfe for x in b.difference_elements
                )

    def initialize(
        self,
        state_args=None,
        outlvl=idaeslog.NOTSET,
        optarg=None,
        solver=None,
        hold_state=True,
        initialize_guess=None,
        type=None,
    ):
        """
        Initialization routine for the membrane channel control volume

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output log level of initialization routine
            optarg : solver options dictionary object (default=None, use
                     default solver options)
            solver : str indicating which solver to use during
                     initialization (default = None)
            hold_state : flag indicating whether the initialization routine
                     should unfix any state variables fixed during
                     initialization, **default** - True. **Valid values:**
                     **True** - states variables are not unfixed, and a dict of
                     returned containing flags for which states were fixed
                     during initialization, **False** - state variables are
                     unfixed after initialization by calling the release_state
                     method.
            initialize_guess : a dict of guesses
        Returns:
            If hold_states is True, returns a dict containing flags for which
            states were fixed during initialization.
        """

        # Get inlet state if not provided
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="control_volume")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="control_volume")

        state_args = self._get_state_args(initialize_guess, state_args)
        state_args_properties_in = state_args["inlet"]

        source_flags = super().initialize(
            state_args=state_args_properties_in,
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            hold_state=True,
        )

        # Differentiate between hot and cold channels for properties_out
        if type == "hot_ch":
            state_args_properties_out = state_args["hot_outlet"]

        elif type == "cold_ch":
            state_args_properties_out = state_args["cold_outlet"]
        else:
            raise ConfigurationError(
                "Either hot_ch or cold_ch must be set in the configuration."
            )

        if hasattr(self, "properties_interface"):
            state_args_interface = self._get_state_args_interface(
                state_args_properties_in, state_args_properties_out
            )

            state_args_interface = self._get_state_args_interface(
                state_args_properties_in, state_args_properties_out
            )
            self.properties_interface.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_args=state_args_interface,
            )

        if hasattr(self, "properties_vapor"):
            state_args_vapor = self._get_state_args_vapor(
                state_args_properties_in, state_args_properties_out
            )

            state_args_vapor = self._get_state_args_vapor(
                state_args_properties_in, state_args_properties_out
            )
            self.properties_vapor.initialize(
                outlvl=outlvl,
                optarg=optarg,
                solver=solver,
                state_args=state_args_vapor,
            )

        init_log.info("Initialization Complete")

        if hold_state:
            return source_flags
        else:
            self.release_state(source_flags, outlvl)

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()

        if hasattr(self, "area"):
            if iscale.get_scaling_factor(self.area) is None:
                iscale.set_scaling_factor(self.area, 100)

        if hasattr(self, "deltaP_channel"):
            for v in self.deltaP_channel.values():
                if iscale.get_scaling_factor(v) is None:
                    iscale.set_scaling_factor(v, 1e-4)

        if hasattr(self, "dP_dx"):
            for v in self.pressure_dx.values():
                iscale.set_scaling_factor(v, 1e-5)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
Tests for zero-order Ozonation model
"""
import pytest


from pyomo.environ import (
    Block,
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    value,
    Var,
)
from pyomo.util.check_units import assert_units_consistent

from idaes.core import FlowsheetBlock
from idaes.core.util.exceptions import ConfigurationError
from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import degrees_of_freedom
from idaes.core.util.testing import initialization_tester
from idaes.core import UnitModelCostingBlock

from watertap.unit_models.zero_order import OzoneZO
from watertap.core.wt_database import Database
from watertap.core.zero_order_properties import WaterParameterBlock
from watertap.costing.zero_order_costing import ZeroOrderCosting

solver = get_solver()


class TestOzoneZO_with_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "cryptosporidium",
                "toc",
                "giardia_lamblia",
                "eeq",
                "total_coliforms_fecal_ecoli",
                "viruses_enteric",
                "tss",
            ]
        )

        m.fs.unit = OzoneZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "giardia_lamblia"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "tss"].fix(1)

        return m

    @pytest.mark.unit
    def test_toc_in_solute_list(self):
        model = ConcreteModel()
        model.db = Database()

        model.fs = FlowsheetBlock(dynamic=False)
        model.fs.params = WaterParameterBlock(
            solute_list=["cryptosporidium", "giardia_lamblia", "eeq"]
        )
        with pytest.raises(
            ConfigurationError,
            match="toc must be in solute list for Ozonation " "or Ozone/AOP",
        ):
            model.fs.unit = OzoneZO(property_package=model.fs.params, database=model.db)

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "ozonation"
        assert isinstance(model.fs.unit.contact_time, Var)
        assert isinstance(model.fs.unit.concentration_time, Var)
        assert isinstance(model.fs.unit.mass_transfer_efficiency, Var)
        assert isinstance(model.fs.unit.ozone_flow_mass, Var)
        assert isinstance(model.fs.unit.ozone_consumption, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.specific_energy_coeff, Var)
        assert isinstance(model.fs.unit.ozone_consumption_constraint, Constraint)
        assert isinstance(model.fs.unit.ozone_flow_mass_constraint, Constraint)
        assert isinstance(model.fs.unit.electricity_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("ozonation")
        model.fs.unit.load_parameters_from_database(use_default_removal=True)
        assert model.fs.unit.recovery_frac_mass_H2O[0].fixed
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"]:
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.contact_time[0].fixed
        assert model.fs.unit.contact_time[0].value == data["contact_time"]["value"]
        assert model.fs.unit.concentration_time[0].fixed
        assert (
            model.fs.unit.concentration_time[0].value
            == data["concentration_time"]["value"]
        )
        assert model.fs.unit.mass_transfer_efficiency[0].fixed
        assert (
            model.fs.unit.mass_transfer_efficiency[0].value
            == data["mass_transfer_efficiency"]["value"]
        )
        assert model.fs.unit.specific_energy_coeff[0].fixed
        assert (
            model.fs.unit.specific_energy_coeff[0].value
            == data["specific_energy_coeff"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.102089, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(2.661333, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(9.795299, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["tss"]
        )
        assert pytest.approx(0.103497, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(9921.863324, rel=1e-5) == value(
            model.fs.unit.ozone_flow_mass[0]
        )
        assert pytest.approx(49609.316620, rel=1e-5) == value(
            model.fs.unit.electricity[0]
        )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


class TestOzoneZO_w_o_default_removal:
    @pytest.fixture(scope="class")
    def model(self):
        m = ConcreteModel()
        m.db = Database()

        m.fs = FlowsheetBlock(dynamic=False)
        m.fs.params = WaterParameterBlock(
            solute_list=[
                "cryptosporidium",
                "toc",
                "giardia_lamblia",
                "eeq",
                "total_coliforms_fecal_ecoli",
                "viruses_enteric",
            ]
        )

        m.fs.unit = OzoneZO(property_package=m.fs.params, database=m.db)

        m.fs.unit.inlet.flow_mass_comp[0, "H2O"].fix(100)
        m.fs.unit.inlet.flow_mass_comp[0, "cryptosporidium"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "toc"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "giardia_lamblia"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "eeq"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "total_coliforms_fecal_ecoli"].fix(1)
        m.fs.unit.inlet.flow_mass_comp[0, "viruses_enteric"].fix(1)

        return m

    @pytest.mark.unit
    def test_build(self, model):
        assert model.fs.unit.config.database is model.db
        assert model.fs.unit._tech_type == "ozonation"
        assert isinstance(model.fs.unit.contact_time, Var)
        assert isinstance(model.fs.unit.concentration_time, Var)
        assert isinstance(model.fs.unit.mass_transfer_efficiency, Var)
        assert isinstance(model.fs.unit.ozone_flow_mass, Var)
        assert isinstance(model.fs.unit.ozone_consumption, Var)
        assert isinstance(model.fs.unit.electricity, Var)
        assert isinstance(model.fs.unit.specific_energy_coeff, Var)
        assert isinstance(model.fs.unit.ozone_consumption_constraint, Constraint)
        assert isinstance(model.fs.unit.ozone_flow_mass_constraint, Constraint)
        assert isinstance(model.fs.unit.electricity_constraint, Constraint)

    @pytest.mark.component
    def test_load_parameters(self, model):
        data = model.db.get_unit_operation_parameters("ozonation")
        model.fs.unit.load_parameters_from_database()
        assert model.fs.unit.recovery_frac_mass_H2O[0].value == 1

        for (t, j), v in model.fs.unit.removal_frac_mass_comp.items():
            assert v.fixed
            if j not in data["removal_frac_mass_comp"]:
                assert v.value == data["default_removal_frac_mass_comp"]["value"]
            else:
                assert v.value == data["removal_frac_mass_comp"][j]["value"]

        assert model.fs.unit.contact_time[0].fixed
        assert model.fs.unit.contact_time[0].value == data["contact_time"]["value"]
        assert model.fs.unit.concentration_time[0].fixed
        assert (
            model.fs.unit.concentration_time[0].value
            == data["concentration_time"]["value"]
        )
        assert model.fs.unit.mass_transfer_efficiency[0].fixed
        assert (
            model.fs.unit.mass_transfer_efficiency[0].value
            == data["mass_transfer_efficiency"]["value"]
        )
        assert model.fs.unit.specific_energy_coeff[0].fixed
        assert (
            model.fs.unit.specific_energy_coeff[0].value
            == data["specific_energy_coeff"]["value"]
        )

    @pytest.mark.component
    def test_degrees_of_freedom(self, model):
        assert degrees_of_freedom(model.fs.unit) == 0

    @pytest.mark.component
    def test_unit_consistency(self, model):
        assert_units_consistent(model.fs.unit)

    @pytest.mark.component
    def test_initialize(self, model):
        initialization_tester(model)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, model):
        results = solver.solve(model)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, model):
        assert pytest.approx(0.101089, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].flow_vol
        )
        assert pytest.approx(2.687660, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["toc"]
        )
        assert pytest.approx(0.957178, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["giardia_lamblia"]
        )
        assert pytest.approx(0.104520, rel=1e-5) == value(
            model.fs.unit.properties_treated[0].conc_mass_comp["eeq"]
        )
        assert pytest.approx(9921.85340, rel=1e-5) == value(
            model.fs.unit.ozone_flow_mass[0]
        )
        assert pytest.approx(49609.267016, rel=1e-5) == value(
            model.fs.unit.electricity[0]
        )

    @pytest.mark.component
    def test_report(self, model):

        model.fs.unit.report()


def test_costing():
    m = ConcreteModel()
    m.db = Database()

    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.params = WaterParameterBlock(solute_list=["sulfur", "toc", "tss"])

    m.fs.costing = ZeroOrderCosting()

    m.fs.unit1 = OzoneZO(property_package=m.fs.params, database=m.db)

    m.fs.unit1.inlet.flow_mass_comp[0, "H2O"].fix(10000)
    m.fs.unit1.inlet.flow_mass_comp[0, "sulfur"].fix(1)
    m.fs.unit1.inlet.flow_mass_comp[0, "toc"].fix(2)
    m.fs.unit1.inlet.flow_mass_comp[0, "tss"].fix(3)
    m.fs.unit1.load_parameters_from_database(use_default_removal=True)
    assert degrees_of_freedom(m.fs.unit1) == 0

    m.fs.unit1.costing = UnitModelCostingBlock(flowsheet_costing_block=m.fs.costing)

    assert isinstance(m.fs.costing.ozonation, Block)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_a_parameter, Var)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_b_parameter, Var)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_c_parameter, Var)
    assert isinstance(m.fs.costing.ozonation.ozone_capital_d_parameter, Var)

    assert isinstance(m.fs.unit1.costing.capital_cost, Var)
    assert isinstance(m.fs.unit1.costing.capital_cost_constraint, Constraint)

    assert_units_consistent(m.fs)
    assert degrees_of_freedom(m.fs.unit1) == 0

    assert m.fs.unit1.electricity[0] in m.fs.costing._registered_flows["electricity"]



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
"""
General costing package for zero-order processes.
"""
import os
import yaml

import pyomo.environ as pyo
from pyomo.common.config import ConfigValue
from pyomo.util.calc_var_value import calculate_variable_from_constraint

from idaes.core import declare_process_block_class
from idaes.core.base.costing_base import register_idaes_currency_units
from watertap.costing.watertap_costing_package import WaterTAPCostingDetailedData

# NOTE: some of these are defined in WaterTAPCostingBlockData
global_params = [
    "plant_lifetime",
    "utilization_factor",
    "land_cost_percent_FCI",
    "working_capital_percent_FCI",
    "salaries_percent_FCI",
    "benefit_percent_of_salary",
    "maintenance_costs_percent_FCI",
    "laboratory_fees_percent_FCI",
    "insurance_and_taxes_percent_FCI",
    "wacc",
    "TPEC",
    "TIC",
]


@declare_process_block_class("ZeroOrderCosting")
class ZeroOrderCostingData(WaterTAPCostingDetailedData):
    """
    General costing package for zero-order processes.
    """

    CONFIG = WaterTAPCostingDetailedData.CONFIG()
    CONFIG.declare(
        "case_study_definition",
        ConfigValue(
            default=None,
            doc="Path to YAML file defining global parameters for case study. If "
            "not provided, default values from the WaterTap database are used.",
        ),
    )

    def register_currency_definitions(self):
        # Register currency and conversion rates
        if "currency_definitions" in self._cs_def:
            pyo.units.load_definitions_from_strings(
                self._cs_def["currency_definitions"]
            )
        else:
            register_idaes_currency_units()

    def set_base_currency_base_period(self):
        # Set the base year for all costs
        self.base_currency = getattr(pyo.units, self._cs_def["base_currency"])
        # Set a base period for all operating costs
        self.base_period = getattr(pyo.units, self._cs_def["base_period"])

    def build_global_params(self):
        """
        To minimize overhead, only create global parameters for now.
        Unit-specific parameters will be added as sub-Blocks on a case-by-case
        basis as a unit of that type is costed.
        """
        # Load case study definition from file
        self._cs_def = _load_case_study_definition(self)

        super().build_global_params()

        # Define expected flows
        for f, v in self._cs_def["defined_flows"].items():
            value = v["value"]
            units = getattr(pyo.units, v["units"])
            if self.component(f + "_cost") is not None:
                self.component(f + "_cost").fix(value * units)
            else:
                self.defined_flows[f] = value * units

        # Fix all Vars from database
        for v in global_params:
            try:
                value = self._cs_def["global_parameters"][v]["value"]
                units = self._cs_def["global_parameters"][v]["units"]
                getattr(self, v).fix(value * getattr(pyo.units, units))
            except KeyError:
                raise KeyError(
                    f"Invalid case study definition file - no entry found "
                    f"for {v}, or entry lacks value and units."
                )
        # this variable floats in this package, so initialize it appropriately here
        calculate_variable_from_constraint(
            self.capital_recovery_factor,
            self.capital_recovery_factor_constraint,
        )


def _load_case_study_definition(self):
    """
    Load data from case study definition file into a Python dict.
    If users did not provide a definition file as a config argument, the
    default definition from the WaterTap techno-economic database is used.
    """
    source_file = self.config.case_study_definition
    if source_file is None:
        source_file = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "..",
            "data",
            "techno_economic",
            "default_case_study.yaml",
        )

    try:
        with open(source_file, "r") as f:
            lines = f.read()
            f.close()
    except OSError:
        raise OSError(
            "Could not find specified case study definition file. "
            "Please check the path provided."
        )

    return yaml.load(lines, yaml.Loader)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

import pytest
from pyomo.environ import (
    ConcreteModel,
    value,
    assert_optimal_termination,
)
from idaes.core import (
    FlowsheetBlock,
)
from watertap.unit_models.osmotically_assisted_reverse_osmosis_1D import (
    OsmoticallyAssistedReverseOsmosis1D,
)
from watertap.unit_models.reverse_osmosis_base import TransportModel
import watertap.property_models.NaCl_prop_pack as props

from watertap.core.solvers import get_solver
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
)
from idaes.core.util.testing import initialization_tester
from idaes.core.util.scaling import (
    calculate_scaling_factors,
)

from watertap.core import (
    ConcentrationPolarizationType,
    MassTransferCoefficient,
    PressureChangeType,
)

from watertap.unit_models.tests.unit_test_harness import UnitTestHarness

# -----------------------------------------------------------------------------
# Get default solver for testing
solver = get_solver()


# -----------------------------------------------------------------------------
def build():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.fixed,
        mass_transfer_coefficient=MassTransferCoefficient.none,
        has_full_reporting=True,
    )

    # fully specify system
    feed_flow_mass = 5 / 18
    feed_mass_frac_NaCl = 0.075
    feed_pressure = 65e5
    feed_temperature = 273.15 + 25
    membrane_area = 155
    width = 1.1
    A = 1e-12
    B = 7.7e-8

    feed_cp_mod = 1.05
    permeate_cp_mod = 0.9

    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.feed_side.cp_modulus.fix(feed_cp_mod)
    m.fs.unit.feed_side.deltaP_stage.fix(0)

    permeate_flow_mass = 0.33 * feed_flow_mass
    permeate_mass_frac_NaCl = 0.1
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        permeate_flow_mass * permeate_mass_frac_H2O
    )
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        permeate_flow_mass * permeate_mass_frac_NaCl
    )
    m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
    m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.permeate_side.cp_modulus.fix(permeate_cp_mod)
    m.fs.unit.permeate_side.deltaP_stage.fix(0)

    m.fs.unit.area.fix(membrane_area)
    m.fs.unit.width.fix(width)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)

    # scaling
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    calculate_scaling_factors(m)

    return m


class TestOsmoticallyAssistedReverse1smosis(UnitTestHarness):
    def configure(self):
        m = build()
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            8.0587e-4
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            4.0855e-6
        )
        self.unit_solutions[
            m.fs.unit.feed_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ] = 0.1320
        self.unit_solutions[
            m.fs.unit.feed_outlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        ] = 0.0202
        self.unit_solutions[m.fs.unit.feed_side.deltaP_stage[0]] = 0
        self.unit_solutions[m.fs.unit.permeate_side.deltaP_stage[0]] = 0

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.feed_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 2": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, 1
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"],
            },
            "Check 3": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.permeate_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 4": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, 1.0
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.permeate_side.cp_modulus[0, 1.0, "NaCl"],
            },
        }

        return m


def build_SKK():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.fixed,
        mass_transfer_coefficient=MassTransferCoefficient.none,
        transport_model=TransportModel.SKK,
        has_full_reporting=True,
    )

    # fully specify system
    feed_flow_mass = 5 / 18
    feed_mass_frac_NaCl = 0.075
    feed_pressure = 65e5
    feed_temperature = 273.15 + 25
    membrane_area = 155
    width = 1.1
    A = 1e-12
    B = 7.7e-8

    feed_cp_mod = 1.05
    permeate_cp_mod = 0.9

    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.feed_side.cp_modulus.fix(feed_cp_mod)
    m.fs.unit.feed_side.deltaP_stage.fix(0)

    permeate_flow_mass = 0.33 * feed_flow_mass
    permeate_mass_frac_NaCl = 0.1
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        permeate_flow_mass * permeate_mass_frac_H2O
    )
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        permeate_flow_mass * permeate_mass_frac_NaCl
    )
    m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
    m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.permeate_side.cp_modulus.fix(permeate_cp_mod)
    m.fs.unit.permeate_side.deltaP_stage.fix(0)

    m.fs.unit.area.fix(membrane_area)
    m.fs.unit.width.fix(width)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.reflect_coeff.fix(0.95)

    # scaling
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    calculate_scaling_factors(m)

    return m


class TestOsmoticallyAssistedReverseOsmosis_SKK(UnitTestHarness):
    def configure(self):
        m = build_SKK()

        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            8.5717e-4
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            9.242e-06
        )
        self.unit_solutions[
            m.fs.unit.feed_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ] = 1.241e-01
        self.unit_solutions[
            m.fs.unit.feed_outlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        ] = 1.940e-02
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[
                0, m.fs.unit.difference_elements.first()
            ].conc_mass_phase_comp["Liq", "NaCl"]
            / m.fs.unit.feed_side.properties[
                0, m.fs.unit.difference_elements.first()
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = value(
            m.fs.unit.feed_side.cp_modulus[
                0, m.fs.unit.difference_elements.first(), "NaCl"
            ]
        )
        self.unit_solutions[m.fs.unit.feed_side.deltaP_stage[0]] = 0
        self.unit_solutions[m.fs.unit.permeate_side.deltaP_stage[0]] = 0

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.feed_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 2": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, 1
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"],
            },
            "Check 3": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.permeate_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 4": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, 1.0
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.permeate_side.cp_modulus[0, 1.0, "NaCl"],
            },
        }

        return m


def build_SKK_CP_calculation_with_kf_calculation():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.calculated,
        transport_model=TransportModel.SKK,
        has_full_reporting=True,
    )

    # fully specify system
    feed_flow_mass = 5 / 18
    feed_mass_frac_NaCl = 0.075
    feed_pressure = 65e5
    feed_temperature = 273.15 + 25
    membrane_area = 155
    width = 1.1
    A = 1e-12
    B = 7.7e-8

    feed_cp_mod = 1.05
    permeate_cp_mod = 0.9

    membrane_pressure_drop = -0.5e-5

    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)
    # m.fs.unit.feed_side.deltaP.fix(membrane_pressure_drop)

    permeate_flow_mass = 0.33 * feed_flow_mass
    permeate_mass_frac_NaCl = 0.1
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        permeate_flow_mass * permeate_mass_frac_H2O
    )
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        permeate_flow_mass * permeate_mass_frac_NaCl
    )
    m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
    m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)

    # m.fs.unit.permeate_side.deltaP.fix(membrane_pressure_drop)

    m.fs.unit.area.fix(membrane_area)
    m.fs.unit.width.fix(width)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)
    m.fs.unit.reflect_coeff.fix(0.95)
    m.fs.unit.structural_parameter.fix(1200e-6)

    m.fs.unit.permeate_side.channel_height.fix(0.002)
    m.fs.unit.permeate_side.spacer_porosity.fix(0.97)
    m.fs.unit.feed_side.channel_height.fix(0.002)
    m.fs.unit.feed_side.spacer_porosity.fix(0.97)

    # scaling
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    calculate_scaling_factors(m)

    return m


class TestOsmoticallyAssistedReverseOsmosis_SKK_CP_calculation_with_kf_calculation(
    UnitTestHarness
):
    def configure(self):
        m = build_SKK_CP_calculation_with_kf_calculation()

        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            7.215e-4
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            8.463e-06
        )
        self.unit_solutions[
            m.fs.unit.feed_outlet.flow_mass_phase_comp[0, "Liq", "H2O"]
        ] = 0.1451
        self.unit_solutions[
            m.fs.unit.feed_outlet.flow_mass_phase_comp[0, "Liq", "NaCl"]
        ] = 0.01952
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 78.8775
        x_interface_in = m.fs.unit.length_domain.at(2)
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 88.360
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 128.619
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[0, 1.0].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 132.351
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 107.06
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[0, 1].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 74.889
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 52.882
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 27.392
        self.unit_solutions[m.fs.unit.feed_side.deltaP_stage[0]] = -197270.178
        self.unit_solutions[m.fs.unit.permeate_side.deltaP_stage[0]] = -109664.455

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.feed_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 2": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, 1
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"],
            },
            "Check 3": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.permeate_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 4": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, 1.0
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.permeate_side.cp_modulus[0, 1.0, "NaCl"],
            },
        }

        return m


def build_SKK_CP_calculation_with_kf_fixed():
    """Testing 1D-OARO with ConcentrationPolarizationType.calculated option enabled.
    This option makes use of an alternative constraint for the feed-side, membrane-interface concentration.
    Additionally, feed and permeate-side mass trasnfer coefficients are fixed.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.fixed,
    )

    # fully specify system
    feed_flow_mass = 5 / 18
    feed_mass_frac_NaCl = 0.075
    feed_pressure = 65e5
    feed_temperature = 273.15 + 25
    membrane_area = 150
    membrane_pressure_drop = -0.5e5
    A = 1e-12
    B = 7.7e-8
    kf = 3.15e-5
    kp = 8.15e-5

    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)

    m.fs.unit.area.fix(membrane_area)

    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)

    permeate_flow_mass = 0.33 * feed_flow_mass
    permeate_mass_frac_NaCl = 0.1
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        permeate_flow_mass * permeate_mass_frac_H2O
    )
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        permeate_flow_mass * permeate_mass_frac_NaCl
    )
    m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
    m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)

    m.fs.unit.feed_side.deltaP.fix(membrane_pressure_drop)
    m.fs.unit.permeate_side.deltaP.fix(membrane_pressure_drop)

    m.fs.unit.feed_side.K[0, 0.0, "NaCl"].fix(kf)
    m.fs.unit.feed_side.K[0, 1.0, "NaCl"].fix(kf)
    m.fs.unit.permeate_side.K[0, 0.0, "NaCl"].fix(kp)
    m.fs.unit.permeate_side.K[0, 1.0, "NaCl"].fix(kp)

    # scaling
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    calculate_scaling_factors(m)

    return m


class TestOsmoticallyAssistedReverseOsmosis_SKK_CP_calculation_with_kf_fixed(
    UnitTestHarness
):
    def configure(self):
        m = build_SKK_CP_calculation_with_kf_fixed()

        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            9.1458e-4
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            3.9757e-6
        )
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 1].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.0825
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 1].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.009167
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 0].flow_mass_phase_comp["Liq", "H2O"]
        ] = 0.2197
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 0].flow_mass_phase_comp["Liq", "NaCl"]
        ] = 0.009763
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 78.8775
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 159.6307
        x_interface_in = m.fs.unit.length_domain.at(2)
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 89.700
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[0, 1.0].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 160.4275
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 107.06
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 43.7056
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 45.1565
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[0, 1.0].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 106.0555
        self.unit_solutions[m.fs.unit.feed_side.cp_modulus[0, 0, "NaCl"]] = 1.1
        self.unit_solutions[m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"]] = 1.005
        self.unit_solutions[m.fs.unit.permeate_side.cp_modulus[0, 0, "NaCl"]] = 1.1
        self.unit_solutions[m.fs.unit.permeate_side.cp_modulus[0, 1, "NaCl"]] = 0.9906

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.feed_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 2": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, 1
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"],
            },
            "Check 3": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.permeate_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 4": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, 1.0
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.permeate_side.cp_modulus[0, 1.0, "NaCl"],
            },
        }

        return m


def build_CP_calculation_with_kf_calculation():
    """Testing 1D-OARO with ConcentrationPolarizationType.calculated option and MassTransferCoefficient.calculated
    option enabled.
    """
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=False,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
    )

    # fully specify system
    feed_flow_mass = 5 / 18
    feed_mass_frac_NaCl = 0.075
    feed_pressure = 65e5
    feed_temperature = 273.15 + 25
    membrane_area = 150
    width = 1.1
    A = 1e-12
    B = 7.7e-8

    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)

    m.fs.unit.area.fix(membrane_area)
    m.fs.unit.width.fix(width)
    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)

    permeate_flow_mass = 0.33 * feed_flow_mass
    permeate_mass_frac_NaCl = 0.1
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        permeate_flow_mass * permeate_mass_frac_H2O
    )
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        permeate_flow_mass * permeate_mass_frac_NaCl
    )
    m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
    m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)

    m.fs.unit.structural_parameter.fix(1200e-6)

    m.fs.unit.permeate_side.channel_height.fix(0.002)
    m.fs.unit.permeate_side.spacer_porosity.fix(0.97)
    m.fs.unit.feed_side.channel_height.fix(0.002)
    m.fs.unit.feed_side.spacer_porosity.fix(0.97)

    # scaling
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    calculate_scaling_factors(m)

    return m


class TestOsmoticallyAssistedReverseOsmosis_CP_calculation_with_kf_calculation(
    UnitTestHarness
):
    def configure(self):
        m = build_CP_calculation_with_kf_calculation()

        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            6.756e-4
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            4.538e-6
        )
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 78.88
        x_interface_in = m.fs.unit.length_domain.at(2)
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 87.056
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[0, 1.0].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 124.028
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[0, 1].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 127.961
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 107.06
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[0, 1.0].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 70.902
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 52.542
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 27.905

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.feed_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 2": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, 1
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"],
            },
            "Check 3": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.permeate_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 4": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, 1.0
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.permeate_side.cp_modulus[0, 1.0, "NaCl"],
            },
        }

        return m


def build_Pdrop_fixed_per_unit_length():
    """Testing 1D-OARO with PressureChangeType.fixed_per_unit_length option."""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.fixed_per_unit_length,
    )

    # fully specify system
    feed_flow_mass = 5 / 18
    feed_mass_frac_NaCl = 0.075
    feed_pressure = 65e5
    feed_temperature = 273.15 + 25
    membrane_area = 150
    A = 1e-12
    B = 7.7e-8
    feed_pressure_drop = -1.9e5
    perm_pressure_drop = -1.7e5
    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.area.fix(membrane_area)

    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)

    permeate_flow_mass = 0.33 * feed_flow_mass
    permeate_mass_frac_NaCl = 0.1
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        permeate_flow_mass * permeate_mass_frac_H2O
    )
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        permeate_flow_mass * permeate_mass_frac_NaCl
    )
    m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
    m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.structural_parameter.fix(1200e-6)

    m.fs.unit.permeate_side.channel_height.fix(0.002)
    m.fs.unit.permeate_side.spacer_porosity.fix(0.97)
    m.fs.unit.feed_side.channel_height.fix(0.002)
    m.fs.unit.feed_side.spacer_porosity.fix(0.97)

    length = 141
    m.fs.unit.length.fix(length)
    m.fs.unit.feed_side.dP_dx.fix(feed_pressure_drop / length)
    m.fs.unit.permeate_side.dP_dx.fix(perm_pressure_drop / length)

    # scaling
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    calculate_scaling_factors(m)

    return m


class TestOsmoticallyAssistedReverseOsmosis_Pdrop_fixed_per_unit_length(
    UnitTestHarness
):
    def configure(self):
        m = build_Pdrop_fixed_per_unit_length()

        self.unit_solutions[m.fs.unit.feed_side.deltaP_stage[0]] = -1.9e5
        self.unit_solutions[m.fs.unit.permeate_side.deltaP_stage[0]] = -1.7e5
        self.unit_solutions[m.fs.unit.feed_side.N_Re[0, 0.0]] = 408.561
        self.unit_solutions[m.fs.unit.feed_side.N_Re[0, 1.0]] = 240.933
        self.unit_solutions[m.fs.unit.permeate_side.N_Re[0, 0.0]] = 298.053
        self.unit_solutions[m.fs.unit.permeate_side.N_Re[0, 1.0]] = 128.761
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            6.740e-4
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            4.513e-6
        )
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 78.878
        x_interface_in = m.fs.unit.length_domain.at(2)
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 87.409
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 123.864
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[0, 1].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 127.557
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 107.06
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[0, 1].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 72.301
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 52.590
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 27.153

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.feed_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 2": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, 1
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"],
            },
            "Check 3": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.permeate_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 4": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, 1.0
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.permeate_side.cp_modulus[0, 1.0, "NaCl"],
            },
        }

        return m


def build_Pdrop_calculation():
    """Testing 1D-OARO with PressureChangeType.calculated option."""
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = props.NaClParameterBlock()

    m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
        property_package=m.fs.properties,
        has_pressure_change=True,
        concentration_polarization_type=ConcentrationPolarizationType.calculated,
        mass_transfer_coefficient=MassTransferCoefficient.calculated,
        pressure_change_type=PressureChangeType.calculated,
    )

    # fully specify system
    feed_flow_mass = 5 / 18
    feed_mass_frac_NaCl = 0.075
    feed_pressure = 65e5
    feed_temperature = 273.15 + 25
    membrane_area = 150
    A = 1e-12
    B = 7.7e-8

    feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        feed_flow_mass * feed_mass_frac_NaCl
    )
    m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        feed_flow_mass * feed_mass_frac_H2O
    )
    m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
    m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.area.fix(membrane_area)

    m.fs.unit.A_comp.fix(A)
    m.fs.unit.B_comp.fix(B)

    permeate_flow_mass = 0.33 * feed_flow_mass
    permeate_mass_frac_NaCl = 0.1
    permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
        permeate_flow_mass * permeate_mass_frac_H2O
    )
    m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
        permeate_flow_mass * permeate_mass_frac_NaCl
    )
    m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
    m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)
    m.fs.unit.structural_parameter.fix(1200e-6)

    m.fs.unit.permeate_side.channel_height.fix(0.002)
    m.fs.unit.permeate_side.spacer_porosity.fix(0.97)
    m.fs.unit.feed_side.channel_height.fix(0.002)
    m.fs.unit.feed_side.spacer_porosity.fix(0.97)
    m.fs.unit.feed_side.N_Re[0, 0].fix(400)

    # scaling
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
    )
    m.fs.properties.set_default_scaling(
        "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
    )
    calculate_scaling_factors(m)

    return m


class TestOsmoticallyAssistedReverseOsmosis_Pdrop_calculation(UnitTestHarness):
    def configure(self):
        m = build_Pdrop_calculation()

        self.unit_solutions[m.fs.unit.feed_side.N_Re[0, 0.0]] = 400
        self.unit_solutions[m.fs.unit.feed_side.velocity[0, 0.0]] = 0.1253
        self.unit_solutions[m.fs.unit.feed_side.N_Re[0, 1.0]] = 237.296
        self.unit_solutions[m.fs.unit.feed_side.velocity[0, 1.0]] = 0.0776
        self.unit_solutions[m.fs.unit.permeate_side.N_Re[0, 0.0]] = 290.356
        self.unit_solutions[m.fs.unit.permeate_side.velocity[0, 0.0]] = 0.0884
        self.unit_solutions[m.fs.unit.permeate_side.N_Re[0, 1.0]] = 126.063
        self.unit_solutions[m.fs.unit.permeate_side.velocity[0, 1.0]] = 0.04062
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "H2O"]] = (
            6.681e-4
        )
        self.unit_solutions[m.fs.unit.flux_mass_phase_comp_avg[0, "Liq", "NaCl"]] = (
            4.471e-6
        )
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 78.878
        x_interface_in = m.fs.unit.length_domain.at(2)
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 87.252
        self.unit_solutions[
            m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 123.235
        self.unit_solutions[
            m.fs.unit.feed_side.properties_interface[0, 1].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 126.981
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 107.06
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[0, 1].conc_mass_phase_comp[
                "Liq", "NaCl"
            ]
        ] = 71.894
        self.unit_solutions[
            m.fs.unit.permeate_side.properties[0, 0].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 52.807
        self.unit_solutions[
            m.fs.unit.permeate_side.properties_interface[
                0, x_interface_in
            ].conc_mass_phase_comp["Liq", "NaCl"]
        ] = 27.618

        self.conservation_equality = {
            "Check 1": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.feed_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 2": {
                "in": m.fs.unit.feed_side.properties_interface[
                    0, 1
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.feed_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.feed_side.cp_modulus[0, 1, "NaCl"],
            },
            "Check 3": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[
                    0, m.fs.unit.difference_elements.first()
                ].conc_mass_phase_comp["Liq", "NaCl"],
                "out": m.fs.unit.permeate_side.cp_modulus[
                    0, m.fs.unit.difference_elements.first(), "NaCl"
                ],
            },
            "Check 4": {
                "in": m.fs.unit.permeate_side.properties_interface[
                    0, 1.0
                ].conc_mass_phase_comp["Liq", "NaCl"]
                / m.fs.unit.permeate_side.properties[0, 1].conc_mass_phase_comp[
                    "Liq", "NaCl"
                ],
                "out": m.fs.unit.permeate_side.cp_modulus[0, 1.0, "NaCl"],
            },
        }

        return m


class TestOsmoticallyAssistedReverseOsmosis_water_recovery:
    water_recovery_list = [0.15, 0.2, 0.3, 0.4, 0.5, 0.55]

    @pytest.mark.parametrize("water_recovery", water_recovery_list)
    @pytest.mark.component
    def test_water_recovery_sweep(self, water_recovery):
        """Testing 1D-OARO with PressureChangeType.calculated option."""
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = props.NaClParameterBlock()

        m.fs.unit = OsmoticallyAssistedReverseOsmosis1D(
            property_package=m.fs.properties,
            has_pressure_change=True,
            concentration_polarization_type=ConcentrationPolarizationType.calculated,
            mass_transfer_coefficient=MassTransferCoefficient.calculated,
            pressure_change_type=PressureChangeType.calculated,
        )

        # fully specify system
        feed_flow_mass = 5 / 18
        feed_mass_frac_NaCl = 0.075
        feed_pressure = 65e5
        feed_temperature = 273.15 + 25
        membrane_area = 150
        A = 1e-12
        B = 7.7e-8

        feed_mass_frac_H2O = 1 - feed_mass_frac_NaCl
        m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
            feed_flow_mass * feed_mass_frac_NaCl
        )
        m.fs.unit.feed_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
            feed_flow_mass * feed_mass_frac_H2O
        )
        m.fs.unit.feed_inlet.pressure[0].fix(feed_pressure)
        m.fs.unit.feed_inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.area.fix(membrane_area)

        m.fs.unit.A_comp.fix(A)
        m.fs.unit.B_comp.fix(B)

        permeate_flow_mass = 0.33 * feed_flow_mass
        permeate_mass_frac_NaCl = 0.1
        permeate_mass_frac_H2O = 1 - permeate_mass_frac_NaCl
        m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "H2O"].fix(
            permeate_flow_mass * permeate_mass_frac_H2O
        )
        m.fs.unit.permeate_inlet.flow_mass_phase_comp[0, "Liq", "NaCl"].fix(
            permeate_flow_mass * permeate_mass_frac_NaCl
        )
        m.fs.unit.permeate_inlet.pressure[0].fix(5e5)
        m.fs.unit.permeate_inlet.temperature[0].fix(feed_temperature)
        m.fs.unit.structural_parameter.fix(1200e-6)

        m.fs.unit.permeate_side.channel_height.fix(0.002)
        m.fs.unit.permeate_side.spacer_porosity.fix(0.97)
        m.fs.unit.feed_side.channel_height.fix(0.002)
        m.fs.unit.feed_side.spacer_porosity.fix(0.97)
        m.fs.unit.feed_side.N_Re[0, 0].fix(400)

        # test degrees of freedom
        assert degrees_of_freedom(m) == 0

        # test scaling
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e1, index=("Liq", "H2O")
        )
        m.fs.properties.set_default_scaling(
            "flow_mass_phase_comp", 1e3, index=("Liq", "NaCl")
        )

        calculate_scaling_factors(m)

        # test initialization
        initialization_tester(m)

        m.fs.unit.permeate_inlet.pressure[0].unfix()
        m.fs.unit.area.unfix()
        m.fs.unit.recovery_mass_phase_comp[0, "Liq", "H2O"].fix(water_recovery)
        m.fs.unit.permeate_outlet.pressure[0].fix(1e5)

        # test solve
        results = solver.solve(m, tee=True)

        # Check for optimal solution
        assert_optimal_termination(results)



#################################################################################
# WaterTAP Copyright (c) 2020-2023, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################
from pyomo.environ import (
    Var,
    Reals,
    Constraint,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock, ConfigValue, In

# Import IDAES cores
from idaes.core import (
    ControlVolume0DBlock,
    declare_process_block_class,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
    UnitModelBlockData,
    useDefault,
    MaterialFlowBasis,
)
from idaes.core.solvers import get_solver
from idaes.core.util.config import is_physical_parameter_block
import idaes.core.util.scaling as iscale
import idaes.logger as idaeslog


__author__ = "Alexander V. Dudchenko"

_log = idaeslog.getLogger(__name__)


@declare_process_block_class("GenericDesalter")
class GenericDesalterData(UnitModelBlockData):
    """
    GenericDesalter - users must provide water recovery
    """

    CONFIG = ConfigBlock()

    CONFIG.declare(
        "dynamic",
        ConfigValue(
            domain=In([False]),
            default=False,
            description="Dynamic model flag - must be False",
            doc="""Indicates whether this model will be dynamic or not,
    **default** = False. NF units do not support dynamic
    behavior.""",
        ),
    )
    CONFIG.declare(
        "has_holdup",
        ConfigValue(
            default=False,
            domain=In([False]),
            description="Holdup construction flag - must be False",
            doc="""Indicates whether holdup terms should be constructed or not.
    **default** - False. NF units do not have defined volume, thus
    this must be False.""",
        ),
    )
    CONFIG.declare(
        "material_balance_type",
        ConfigValue(
            default=MaterialBalanceType.useDefault,
            domain=In(MaterialBalanceType),
            description="Material balance construction flag",
            doc="""Indicates what type of mass balance should be constructed,
    **default** - MaterialBalanceType.useDefault.
    **Valid values:** {
    **MaterialBalanceType.useDefault - refer to property package for default
    balance type
    **MaterialBalanceType.none** - exclude material balances,
    **MaterialBalanceType.componentPhase** - use phase component balances,
    **MaterialBalanceType.componentTotal** - use total component balances,
    **MaterialBalanceType.elementTotal** - use total element balances,
    **MaterialBalanceType.total** - use total material balance.}""",
        ),
    )
    CONFIG.declare(
        "energy_balance_type",
        ConfigValue(
            default=EnergyBalanceType.useDefault,
            domain=In(EnergyBalanceType),
            description="Energy balance construction flag",
            doc="""Indicates what type of energy balance should be constructed,
    **default** - EnergyBalanceType.useDefault.
    **Valid values:** {
    **EnergyBalanceType.useDefault - refer to property package for default
    balance type
    **EnergyBalanceType.none** - exclude energy balances,
    **EnergyBalanceType.enthalpyTotal** - single enthalpy balance for material,
    **EnergyBalanceType.enthalpyPhase** - enthalpy balances for each phase,
    **EnergyBalanceType.energyTotal** - single energy balance for material,
    **EnergyBalanceType.energyPhase** - energy balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "momentum_balance_type",
        ConfigValue(
            default=MomentumBalanceType.pressureTotal,
            domain=In(MomentumBalanceType),
            description="Momentum balance construction flag",
            doc="""Indicates what type of momentum balance should be constructed,
    **default** - MomentumBalanceType.pressureTotal.
    **Valid values:** {
    **MomentumBalanceType.none** - exclude momentum balances,
    **MomentumBalanceType.pressureTotal** - single pressure balance for material,
    **MomentumBalanceType.pressurePhase** - pressure balances for each phase,
    **MomentumBalanceType.momentumTotal** - single momentum balance for material,
    **MomentumBalanceType.momentumPhase** - momentum balances for each phase.}""",
        ),
    )
    CONFIG.declare(
        "has_pressure_change",
        ConfigValue(
            default=False,
            domain=In([True, False]),
            description="Pressure change term construction flag",
            doc="""Indicates whether terms for pressure change should be
    constructed,
    **default** - False.
    **Valid values:** {
    **True** - include pressure change terms,
    **False** - exclude pressure change terms.}""",
        ),
    )
    CONFIG.declare(
        "property_package",
        ConfigValue(
            default=useDefault,
            domain=is_physical_parameter_block,
            description="Property package to use for control volume",
            doc="""Property parameter object used to define property calculations,
    **default** - useDefault.
    **Valid values:** {
    **useDefault** - use default package from parent model or flowsheet,
    **PhysicalParameterObject** - a PhysicalParameterBlock object.}""",
        ),
    )
    CONFIG.declare(
        "property_package_args",
        ConfigBlock(
            implicit=True,
            description="Arguments to use for constructing property packages",
            doc="""A ConfigBlock with arguments to be passed to a property block(s)
    and used when constructing these,
    **default** - None.
    **Valid values:** {
    see property package for documentation.}""",
        ),
    )

    CONFIG.declare(
        "tracked_solids_list",
        ConfigValue(
            default=None,
            domain=list,
            description="List of solids that can form in desalter unit",
            doc="""A list of solids that will should be tracked in the waste of desalter unit, 
            if None, then model will not include calculation of water content in solids or 
            solids concentration""",
        ),
    )

    def build(self):
        # Call UnitModel.build to setup dynamics
        super().build()

        # self.scaling_factor = Suffix(direction=Suffix.EXPO)
        self._get_property_package()
        units_meta = self.config.property_package.get_metadata().get_derived_units

        self.water_recovery = Var(
            initialize=80,
            bounds=(None, 99.9999),
            domain=Reals,
            units=pyunits.dimensionless,
            doc="water recovery",
        )
        self.water_recovery.fix()
        if self.config.tracked_solids_list is not None:
            self.brine_solids_concentration = Var(
                initialize=80,
                bounds=(None, None),
                domain=Reals,
                units=pyunits.kg / pyunits.m**3,
                doc="brine solids concentration",
            )
            self.brine_water_mass_percent = Var(
                initialize=80,
                bounds=(None, None),
                domain=Reals,
                units=pyunits.dimensionless,
                doc="brine water mass percent",
            )
        self.brine_unit = ControlVolume0DBlock(
            dynamic=False,
            has_holdup=False,
            property_package=self.config.property_package,
            property_package_args=self.config.property_package_args,
        )
        self.brine_unit.add_state_blocks(has_phase_equilibrium=False)

        self.brine_unit.add_material_balances(
            balance_type=self.config.material_balance_type,
            has_mass_transfer=True,
        )
        self.product_properties = self.config.property_package.build_state_block(
            self.flowsheet().time,
            doc="Material properties in feed",
            defined_state=True,
            has_phase_equilibrium=False,
            **self.config.property_package_args
        )
        self.product_properties[0].define_state_vars()
        self.add_inlet_port(
            name="inlet",
            block=self.brine_unit.properties_in,
        )
        self.add_port(name="brine", block=self.brine_unit.properties_out)
        self.add_port(name="product", block=self.product_properties)

        @self.brine_unit.Constraint(
            self.flowsheet().config.time,
            doc="isothermal energy balance for reactor",
        )
        def eq_isothermal_brine_unit(b, t):
            return b.properties_in[t].temperature == b.properties_out[t].temperature

        @self.brine_unit.Constraint(
            self.flowsheet().config.time,
            doc="isothermal energy balance for reactor",
        )
        def eq_isobaric_brine_unit(b, t):
            return b.properties_in[t].pressure == b.properties_out[t].pressure

        @self.Constraint(
            self.flowsheet().config.time,
            doc="isothermal energy balance for reactor",
        )
        def eq_isothermal(b, t):
            return (
                b.brine_unit.properties_in[t].temperature
                == b.product_properties[t].temperature
            )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="isothermal energy balance for reactor",
        )
        def eq_isobaric(b, t):
            return (
                b.brine_unit.properties_in[t].pressure
                == b.product_properties[t].pressure
            )

        @self.brine_unit.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass balance",
        )
        def eq_ion_transfer(b, t, phase, ion):
            if ion == "H2O":
                return Constraint.Skip
            else:
                if (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.mass
                ):
                    return (
                        b.properties_in[t].flow_mass_phase_comp[phase, ion]
                        == b.properties_out[t].flow_mass_phase_comp[phase, ion]
                    )
                elif (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.molar
                ):
                    return (
                        b.properties_in[t].flow_mol_phase_comp[phase, ion]
                        == b.properties_out[t].flow_mol_phase_comp[phase, ion]
                    )

        @self.Constraint(
            self.flowsheet().config.time,
            self.config.property_package.phase_list,
            self.config.property_package.component_list,
            doc="Mass balance",
        )
        def eq_product_ion_transfer(b, t, phase, ion):
            if ion == "H2O":
                return Constraint.Skip
            else:
                if (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.mass
                ):
                    return (
                        b.product_properties[t].flow_mass_phase_comp[phase, ion] == 0.0
                    )
                elif (
                    b.config.property_package.config.material_flow_basis
                    == MaterialFlowBasis.molar
                ):
                    return (
                        b.product_properties[t].flow_mass_phase_comp[phase, ion] == 0.0
                    )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Mass balance with reaction terms",
        )
        def eq_water_recovery(b, t):
            return (
                b.brine_unit.properties_in[t].flow_vol_phase["Liq"]
                * (self.water_recovery / 100)
                == b.product_properties[t].flow_vol_phase["Liq"]
            )

        @self.Constraint(
            self.flowsheet().config.time,
            doc="Mass balance with reaction terms",
        )
        def eq_brine_flow(b, t):
            return b.brine_unit.properties_out[t].flow_vol_phase[
                "Liq"
            ] == b.brine_unit.properties_in[t].flow_vol_phase["Liq"] * (
                1 - self.water_recovery / 100
            )

        if self.config.tracked_solids_list is not None:

            @self.Constraint(
                self.flowsheet().config.time,
                doc="Mass balance with reaction terms",
            )
            def solids_concentration_eq(b, t):
                return b.brine_solids_concentration * b.brine_unit.properties_out[
                    t
                ].flow_vol_phase["Liq"] == sum(
                    [
                        b.brine_unit.properties_out[t].flow_mass_phase_comp["Liq", tds]
                        for tds in self.config.tracked_solids_list
                    ]
                )

            @self.Constraint(
                self.flowsheet().config.time,
                doc="Mass balance with reaction terms",
            )
            def water_mass_frac_eq(b, t):
                return (
                    b.brine_water_mass_percent
                    * (
                        sum(
                            [
                                b.brine_unit.properties_out[0].flow_mass_phase_comp[
                                    "Liq", tds
                                ]
                                for tds in self.config.tracked_solids_list
                            ]
                        )
                        + b.brine_unit.properties_out[0].flow_mass_phase_comp[
                            "Liq", "H2O"
                        ]
                    )
                    == b.brine_unit.properties_out[0].flow_mass_phase_comp["Liq", "H2O"]
                    * 100
                )

    def initialize_build(
        self, state_args=None, outlvl=idaeslog.NOTSET, solver=None, optarg=None
    ):
        """
        General wrapper for initialization routines

        Keyword Arguments:
            state_args : a dict of arguments to be passed to the property
                         package(s) to provide an initial state for
                         initialization (see documentation of the specific
                         property package) (default = {}).
            outlvl : sets output level of initialization routine
            optarg : solver options dictionary object (default=None)
            solver : str indicating which solver to use during
                     initialization (default = None)

        Returns: None
        """
        init_log = idaeslog.getInitLogger(self.name, outlvl, tag="unit")
        solve_log = idaeslog.getSolveLogger(self.name, outlvl, tag="unit")

        opt = get_solver(solver, optarg)
        control_vol_flags = self.brine_unit.initialize(
            outlvl=outlvl,
            optarg=optarg,
            solver=solver,
            state_args=state_args,
        )
        init_log.info_high("Dissolution reactor Step 1 Complete.")
        with idaeslog.solver_log(solve_log, idaeslog.DEBUG) as slc:
            res = opt.solve(self, tee=slc.tee)
        init_log.info_high("Initialization Step 3 {}.".format(idaeslog.condition(res)))
        self.brine_unit.release_state(control_vol_flags, outlvl)
        init_log.info("Initialization Complete: {}".format(idaeslog.condition(res)))

    def calculate_scaling_factors(self):
        super().calculate_scaling_factors()
        for (t, p, j), con in self.brine_unit.eq_ion_transfer.items():
            sf = iscale.get_scaling_factor(
                self.brine_unit.properties_in[t].get_material_flow_terms(p, j)
            )
            iscale.constraint_scaling_transform(con, sf)
        sf = iscale.get_scaling_factor(
            self.brine_unit.properties_in[0].pressure, 1 / 1e5
        )
        iscale.constraint_scaling_transform(self.eq_isobaric[0], sf)
        iscale.constraint_scaling_transform(
            self.brine_unit.eq_isobaric_brine_unit[0], sf
        )
        sf = iscale.get_scaling_factor(
            self.brine_unit.properties_in[0].temperature, 1 / 100
        )
        iscale.constraint_scaling_transform(
            self.brine_unit.eq_isothermal_brine_unit[0], sf
        )
        iscale.constraint_scaling_transform(self.eq_isothermal[0], sf)
        iscale.constraint_scaling_transform(self.eq_water_recovery[0], 1)
        iscale.set_scaling_factor(self.water_recovery, 1)



#################################################################################
# WaterTAP Copyright (c) 2020-2024, The Regents of the University of California,
# through Lawrence Berkeley National Laboratory, Oak Ridge National Laboratory,
# National Renewable Energy Laboratory, and National Energy Technology
# Laboratory (subject to receipt of any required approvals from the U.S. Dept.
# of Energy). All rights reserved.
#
# Please see the files COPYRIGHT.md and LICENSE.md for full copyright and license
# information, respectively. These files are also available online at the URL
# "https://github.com/watertap-org/watertap/"
#################################################################################

__author__ = "Paul Vecchiarelli, Ben Knueven, Adam Atia"

from collections import UserDict
from pyomo.environ import units as pyunits
from pyomo.core.base.units_container import _PyomoUnit
from collections.abc import Iterable


class FixedKeysDict(UserDict):
    def __init__(self, d):
        self.data = d

    def __setitem__(self, k, v):
        if k not in self.data:
            raise RuntimeError(f" Key {k} not in dictionary.")
        # also check for valid value if a list of values is given in the default dictionary
        else:
            # if user setting value in pyomo units
            if hasattr(v, "is_expression_type"):
                if isinstance(v, _PyomoUnit) or v.is_expression_type():
                    # if user assigns pyomo units as value to oli_unit, save the str to oli_unit and update pyomo_unit with pyomo units
                    if "oli_unit" in k:
                        self.data[k] = str(v)
                        self.data["pyomo_unit"] = v
                    # if user assigns pyomo units to pyomo_unit, update oli_unit with str representation of units
                    if "pyomo_unit" in k:
                        self.data[k] = v
                        self.data["oli_unit"] = str(v)
            # check if data[k] is iterable first, otherwise checking if oli_unit in data[k] throws exception
            elif isinstance(self.data[k], Iterable):
                # check if user provides str and that assignment wouldn't overwrite the oli_unit key:value pair
                if isinstance(v, str) and ("oli_unit" not in self.data[k]):
                    # if user assigns str to oli_unit, update pyomo_units with PyomoUnits representation of str
                    if "oli_unit" in k:
                        self.data[k] = v
                        self.data["pyomo_unit"] = getattr(pyunits, v)
                    else:
                        pass
                elif isinstance(v, str) and ("oli_unit" in self.data[k]):
                    self.data["oli_unit"] = v
                    self.data["pyomo_unit"] = getattr(pyunits, v)
                elif not isinstance(v, str):
                    raise RuntimeError(
                        f"Setting {v} as the value for {k} is not permitted as a value for oli and pyomo units. Please enter units as a string type or pint units."
                    )
                else:
                    pass
            elif not isinstance(self.data[k], Iterable):
                if isinstance(v, str) and "pyomo_unit" in k:
                    self.data[k] = getattr(pyunits, v)
                    self.data["oli_unit"] = v
                elif not isinstance(v, str):
                    raise RuntimeError(
                        f"Setting {v} as the value for {k} is not permitted as a value for oli and pyomo units. Please enter units as a string type or pint units."
                    )
                else:
                    pass
            else:
                self.data[k] = v

    def __delitem__(self, k):
        raise Exception(" Deleting keys not supported for this object.")

    def _check_value(self, k, valid_values):
        if k not in self.data:
            raise RuntimeError(f" Key {k} not in dictionary.")
        else:
            if self.data[k] not in valid_values:
                raise RuntimeError(
                    f" Value {self.data[k]} not a valid value for key {k}."
                )

    def pprint(self):
        print("-------------------------")
        for key, value in self.data.items():
            print(f" {key}\n - {value}\n")


input_unit_set_temp = {
    "inflows": {
        "oli_unit": "mg/L",
        "pyomo_unit": pyunits.mg / pyunits.L,
    },
    "molecularConcentration": {
        "oli_unit": "mg/L",
        "pyomo_unit": pyunits.mg / pyunits.L,
    },
    "mass": {"oli_unit": "mg", "pyomo_unit": pyunits.mg},
    "temperature": {"oli_unit": "K", "pyomo_unit": pyunits.K},
    "pressure": {"oli_unit": "Pa", "pyomo_unit": pyunits.Pa},
    "enthalpy": {"oli_unit": "J", "pyomo_unit": pyunits.J},
    "vaporAmountMoles": {"oli_unit": "mol", "pyomo_unit": pyunits.mol},
    "vaporMolFrac": {
        "oli_unit": "mol/mol",
        "pyomo_unit": pyunits.mol / pyunits.mol,
    },
    "totalVolume": {"oli_unit": "L", "pyomo_unit": pyunits.L},
    "pipeDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "pipeFlowVelocity": {
        "oli_unit": "m/s",
        "pyomo_unit": pyunits.meter / pyunits.second,
    },
    "diskDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "diskRotatingSpeed": {"oli_unit": "cycle/s", "pyomo_unit": 1 / pyunits.second},
    "rotorDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "rotorRotation": {"oli_unit": "cycle/s", "pyomo_unit": 1 / pyunits.second},
    "shearStress": {"oli_unit": "Pa", "pyomo_unit": pyunits.Pa},
    "pipeDiameter": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "pipeRoughness": {"oli_unit": "m", "pyomo_unit": pyunits.meter},
    "liquidFlowInPipe": {
        "oli_unit": "L/s",
        "pyomo_unit": pyunits.L / pyunits.second,
    },
    "gasFlowInPipe": {"oli_unit": "L/s", "pyomo_unit": pyunits.L / pyunits.second},
    "viscAbs2ndLiq": {
        "oli_unit": "Pa-s",
        "pyomo_unit": pyunits.Pa * pyunits.second,
    },
    "alkalinity": {"oli_unit": "mg HCO3/L", "pyomo_unit": pyunits.mg / pyunits.L},
    "TIC": {"oli_unit": "mol C/L", "pyomo_unit": pyunits.mol / pyunits.L},
    "CO2GasFraction": {
        "oli_unit": "mol/mol",
        "pyomo_unit": pyunits.mol / pyunits.mol,
    },
}


# TODO: consider adding these: https://devdocs.olisystems.com/user-defined-output-unit-set
# and reducing hard-coding by using default_input_unit_set references
output_unit_set_temp = {
    "enthalpy": input_unit_set_temp["enthalpy"],
    "mass": input_unit_set_temp["mass"],
    "pt": input_unit_set_temp["pressure"],
    "total": input_unit_set_temp["mass"],
    "liq1_phs_comp": input_unit_set_temp["mass"],
    "solid_phs_comp": input_unit_set_temp["mass"],
    "vapor_phs_comp": input_unit_set_temp["mass"],
    "liq2_phs_comp": input_unit_set_temp["mass"],
    "combined_phs_comp": input_unit_set_temp["mass"],
    "molecularConcentration": input_unit_set_temp["molecularConcentration"],
}

input_unit_set = FixedKeysDict(
    {k: FixedKeysDict(v) for k, v in input_unit_set_temp.items()}
)
output_unit_set = FixedKeysDict(
    {k: FixedKeysDict(v) for k, v in output_unit_set_temp.items()}
)

optional_properties = FixedKeysDict(
    {
        "electricalConductivity": True,
        "viscosity": True,
        "selfDiffusivityAndMobility": True,
        "heatCapacity": True,
        "thermalConductivity": True,
        "surfaceTension": True,
        "interfacialTension": True,
        "volumeStdConditions": True,
        "prescalingTendenciesEstimated": False,
        "prescalingIndexEstimated": False,
        "prescalingTendenciesRigorous": True,
        "prescalingIndexRigorous": True,
        "scalingTendencies": True,
        "scalingIndex": True,
        "hardness": True,
        "ionicStrengthXBased": True,
        "ionicStrengthMBased": True,
        "totalDissolvedSolids": True,
        "vaporToInflowMoleFraction": True,
        "partialPressure": True,
        "vaporDiffusivityMatrix": True,
        "entropyStream": True,
        "entropySpecies": True,
        "entropyStreamStandardState": True,
        "entropySpeciesStandardState": True,
        "gibbsEnergyStream": True,
        "gibbsEnergySpecies": True,
        "gibbsEnergyStreamStandardState": True,
        "gibbsEnergySpeciesStandardState": True,
        "activityCoefficientsXBased": True,
        "activityCoefficientsMBased": True,
        "fugacityCoefficients": True,
        "vaporFugacity": True,
        "kValuesXBased": True,
        "kValuesMBased": True,
        "MBGComposition": True,
        "materialBalanceGroup": True,
    }
)


if __name__ == "__main__":
    unit_set = input_unit_set
    output_unit_set = output_unit_set
